"""
Shapes classes.

Library Version 1.0

Copyright 2013 Paul Griffiths
Email: mail@paulgriffiths.net

Distributed under the terms of the GNU General Public License.
http://www.gnu.org/licenses/
"""

# pylint: disable=R0201

from math import pi
from geometry.conversions import validate_length, validate_area, convert


class _Shape():

    """
    Base shape class.
    """

    def __init__(self, units):

        """
        Class initializer.

        Arguments:
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        """

        validate_length(units)
        self._units = units
        self._units_area = "{0}2".format(units)

    def _area_internal_units(self):

        """
        Virtual method - returns the area of the shape
        in squared internal units.
        """

        return 1

    def _perimeter_internal_units(self):

        """
        Virtual method - returns the perimeter of the shape
        in internal units.
        """

        return 1

    def area(self, areaunits=None):

        """
        Returns the area of the shape in user units.

        Arguments:
        areaunits -- units in which to return area, default is
        squared internal length units
        """

        if not areaunits:
            areaunits = self._units_area
        else:
            validate_area(areaunits)

        area_int = self._area_internal_units()
        return convert(area_int, self._units_area, areaunits)

    def perimeter(self, units=None):

        """
        Returns the perimeter of the shape in user units.

        Arguments:
        units -- units in which to return perimeter, default is
        internal length units
        """

        if not units:
            units = self._units
        else:
            validate_length(units)

        perim_int = self._perimeter_internal_units()
        return convert(perim_int, self._units, units)


class Rectangle(_Shape):

    """
    Rectangle class.
    """

    def __init__(self, sides, units="cm"):

        """
        Class initializer.

        Arguments:
        sides -- a two-element tuple containing the lengths of
        the two sides of the rectangle
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        """

        _Shape.__init__(self, units)

        self._side_a, self._side_b = sides

    def _area_internal_units(self):

        """
        Returns the area of the rectangle in squared internal units.
        """

        return self._side_a * self._side_b

    def _perimeter_internal_units(self):

        """
        Returns the perimeter of the rectangle in squared internal units.
        """

        return (self._side_a + self._side_b) * 2


class Square(Rectangle):

    """
    Square class.
    """

    def __init__(self, side, units="cm"):

        """
        Class initializer.

        Arguments:
        side -- the length of the sides of the square
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        """

        Rectangle.__init__(self, (side, side), units)


class Circle(_Shape):

    """
    Circle class.
    """

    def __init__(self, radius, units="cm"):

        """
        Class initializer.

        Arguments:
        radius -- the radius of the circle
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        """

        _Shape.__init__(self, units)

        self._radius = radius

    def _area_internal_units(self):

        """
        Returns the area of the circle in squared internal units.
        """

        return self._radius ** 2 * pi

    def _perimeter_internal_units(self):

        """
        Returns the perimeter of the circle in squared internal units.
        """

        return self._radius * 2 * pi
