"""
Solids classes.

Library Version 1.0

Copyright 2013 Paul Griffiths
Email: mail@paulgriffiths.net

Distributed under the terms of the GNU General Public License.
http://www.gnu.org/licenses/
"""

# pylint: disable=R0201

from math import pi, sqrt, tan
from geometry.materials import DefaultMaterial
from geometry.conversions import validate_length, validate_area
from geometry.conversions import validate_volume, validate_weight, convert
from geometry.shapes import Rectangle, Circle


class _Solid():

    """
    Base solid class.
    """

    def __init__(self, units, material):

        """
        Class initializer.

        Arguments:
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        validate_length(units)
        self._units = units
        self._units_area = "{0}2".format(units)
        self._units_volume = "{0}3".format(units)

        self._material = material

    def _volume_internal_units(self):

        """
        Virtual method - returns the volume of the solid in internal units.
        """

        return 1

    def _surface_area_internal_units(self):

        """
        Virtual method - returns the surface area of the solid
        in internal units.
        """

        return 1

    def volume(self, volumeunits=None):

        """
        Returns the volume of the cuboid in user units.

        Arguments:
        volumeunits -- units in which to return volume, default is
        cubed internal length units.
        """

        if not volumeunits:
            volumeunits = self._units_volume
        else:
            validate_volume(volumeunits)

        volume_int = self._volume_internal_units()
        return convert(volume_int, self._units_volume, volumeunits)

    def surface_area(self, areaunits=None):

        """
        Returns the surface area of the cuboid in user units.

        Arguments:
        areaunits -- units in which to return surface area, default is
        squared internal length units.
        """

        if not areaunits:
            areaunits = self._units_area
        else:
            validate_area(areaunits)

        area_int = self._surface_area_internal_units()
        return convert(area_int, self._units_area, areaunits)

    def weight(self, weightunits="grams"):

        """
        Virtual method - returns the weight of the solid.
        """

        validate_weight(weightunits)

        return self._material.weight(self.volume("cc"), "cc", weightunits)


class Cuboid(_Solid):

    """
    Cuboid class.
    """

    def __init__(self, edges, units="cm", material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        edges --- a three-element tuple containing the lengths of
        the edges of the cuboid
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        _Solid.__init__(self, units, material)

        self._edge_a, self._edge_b, self._edge_c = edges

    def _volume_internal_units(self):

        """
        Returns the volume of the cuboid in cubic internal units.
        """

        return self._edge_a * self._edge_b * self._edge_c

    def _surface_area_internal_units(self):

        """
        Returns the surface area of the cuboid in squared internal units.
        """

        return (Rectangle((self._edge_a, self._edge_b)).area() +
                Rectangle((self._edge_b, self._edge_c)).area() +
                Rectangle((self._edge_c, self._edge_a)).area()) * 2


class Cube(Cuboid):

    """
    Cube class.
    """

    def __init__(self, edge, units="cm", material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        edge --- the length of the edges of the cube
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        Cuboid.__init__(self, (edge, edge, edge), units, material)


class Sphere(_Solid):

    """
    Sphere class.
    """

    def __init__(self, radius, units="cm", material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        radius --- the radius of the sphere
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        _Solid.__init__(self, units, material)

        self._radius = radius

    def _volume_internal_units(self):

        """
        Returns the volume of the sphere in cubic internal units.
        """

        return self._radius ** 3 * pi * 4 / 3

    def _surface_area_internal_units(self):

        """
        Returns the surface area of the sphere in squared internal units.
        """

        return Circle(self._radius).area() * 4


class Cone(_Solid):

    """
    Cone class.
    """

    def __init__(self, radius, height, units="cm",
                 material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        radius --- the radius of the base of the cone
        height --- the height of the cone
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        _Solid.__init__(self, units, material)

        self._radius = radius
        self._height = height

    def _volume_internal_units(self):

        """
        Returns the volume of the cone in cubic internal units.
        """

        return Circle(self._radius).area() * self._height / 3

    def _surface_area_internal_units(self):

        """
        Returns the surface area of the cone in squared internal units.
        """

        hyp = sqrt(self._radius ** 2 + self._height ** 2)
        return self._radius * pi * hyp + Circle(self._radius).area()


class Cylinder(_Solid):

    """
    Cylinder class.
    """

    def __init__(self, radius, height, units="cm",
                 material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        radius --- the radius of the bases of the cylinder
        height --- the height (or length) of the cylinder
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        _Solid.__init__(self, units, material)

        self._radius = radius
        self._height = height

    def _volume_internal_units(self):

        """
        Returns the volume of the cylinder in cubic internal units.
        """

        return Circle(self._radius).area() * self._height

    def _surface_area_internal_units(self):

        """
        Returns the surface area of the cylinder in squared internal units.
        """

        return (self._lateral_sa_internal_units() +
                self._base_sa_internal_units() * 2)

    def _lateral_sa_internal_units(self):

        """
        Returns the lateral surface area of the cylinder in squared
        internal units.
        """

        return Circle(self._radius).perimeter() * self._height

    def _base_sa_internal_units(self):

        """
        Returns the base surface area of the cylinder in squared
        internal units.
        """

        return Circle(self._radius).area()

    def lateral_surface_area(self, areaunits=None):

        """
        Returns the lateral surface area of the cylinder in user units.

        Arguments:
        areaunits -- units in which to return surface area, default is
        squared internal length units.
        """

        if not areaunits:
            areaunits = self._units_area
        else:
            validate_area(areaunits)

        area_int = self._lateral_sa_internal_units()
        return convert(area_int, self._units_area, areaunits)

    def base_surface_area(self, areaunits=None):

        """
        Returns the base surface area of the cylinder in user units.

        Arguments:
        areaunits -- units in which to return surface area, default is
        squared internal length units.
        """

        if not areaunits:
            areaunits = self._units_area
        else:
            validate_area(areaunits)

        area_int = self._base_sa_internal_units()
        return convert(area_int, self._units_area, areaunits)


class Tube(_Solid):

    """
    Tube class.
    """

    def __init__(self, radii, length, units="cm",
                 material=DefaultMaterial()):

        """
        Class initializer.

        Arguments:
        radii --- a two-element tuple containing firstly the outer
        radius of the tube, and secondly the inner radius of the tube
        length --- the length of the tube
        units -- a string containing the length units for the
        shape, e.g. "mm", "km", "furlong", "mile"
        material -- a material object containing the material
        with which to construct the solid
        """

        _Solid.__init__(self, units, material)

        self._radius_outer, self._radius_inner = radii
        self._length = length

    def _volume_internal_units(self):

        """
        Returns the volume of the tube in cubic internal units.
        """

        return (Cylinder(self._radius_outer, self._length).volume() -
                Cylinder(self._radius_inner, self._length).volume())

    def _surface_area_internal_units(self):

        """
        Returns the surface area of the tube in squared internal units.
        """

        r_o, r_i = self._radius_outer, self._radius_inner
        length = self._length

        end = (Cylinder(r_o, length).base_surface_area() -
               Cylinder(r_i, length).base_surface_area())

        lat = (Cylinder(r_o, length).lateral_surface_area() +
               Cylinder(r_i, length).lateral_surface_area())

        return lat + end * 2

    def segment_extrados(self, nominalradius, angle):

        """
        Returns the extrados dimension of a tube segment of
        a specified angle where those segments would form a
        tube bend with a specified nominal radius.

        Arguments:
        nominalradius -- the nominal radius of the bend
        angle -- the segment angle, in radians
        """

        return self._segment_dimension("extra", nominalradius, angle)

    def segment_intrados(self, nominalradius, angle):

        """
        Returns the intrados dimension of a tube segment of
        a specified angle where those segments would form a
        tube bend with a specified nominal radius.

        Arguments:
        nominalradius -- the nominal radius of the bend
        angle -- the segment angle, in radians
        """

        return self._segment_dimension("inner", nominalradius, angle)

    def segment_mean_length(self, nominalradius, angle):

        """
        Returns the mean length dimension of a tube segment of
        a specified angle where those segments would form a
        tube bend with a specified nominal radius.

        Arguments:
        nominalradius -- the nominal radius of the bend
        angle -- the segment angle, in radians
        """

        return self._segment_dimension("mean", nominalradius, angle)

    def _segment_dimension(self, dim, nominalradius, angle):

        """
        Returns a specified dimension of a tube segment of
        a specified angle where those segments would form a
        tube bend with a specified nominal radius.

        Arguments:
        dim -- the dimension to return, either "extra", "inner", or "mean"
        nominalradius -- the nominal radius of the bend
        angle -- the segment angle, in radians
        """

        if dim == "extra":
            rad = nominalradius + self._radius_outer
        elif dim == "inner":
            rad = nominalradius - self._radius_outer
        else:
            rad = nominalradius

        return rad * tan(angle / 2) * 2
